import time
from typing import Any, Literal, Optional
import requests

from .exception import OMMXDA4AdapterError
from .models import QuboRequest, QuboResponse, JobID


class DA4Client:
    """A Client class for Fujitsu Digital Annealer API (DA4).

    This class provides an interface for interacting with the Digital Annealer API.
    """

    def __init__(
        self,
        *,
        token: str,
        url: str = "https://api.aispf.global.fujitsu.com/da",
        version: Literal["v4", "v3c"] = "v4",
    ):
        """Initialize DA4 Client.

        :param token: Authentication token for DA4 API.
        :param url: URL to the Fujitsu Digital Annealer. Defaults to "https://api.aispf.global.fujitsu.com/da".
        :param version: The version of Digital Annealer as either "v4" or "v3c". Defaults to "v4".
        :raises OMMXDA4AdapterError: If the version is not "v4" or "v3c"
        """
        if version not in set(["v4", "v3c"]):
            raise OMMXDA4AdapterError(
                "The version of Digital Annealer must be either 'v4' or 'v3c'."
            )

        self._token = token
        self._url = url
        self._headers = {
            "accept": "application/json",
            "Content-Type": "application/json",
            "X-Api-Key": self._token,
        }
        self._version = version
        self._api_url_jobs = f"/{self._version}/async/jobs"

        # The max trial count to access the Fujitsu Digital Annealer.
        self._max_loop_count = 10000

        # The access interval [sec] to access the Fujitsu Digital Annealer.
        self._solver_access_interval = 10

    def get(self, url: str) -> dict[str, Any]:
        """wrapper of `requests.get`

        :param url: The endpoint path
        :return: json dict
        :raises OMMXDA4AdapterError: If an HTTP status other than the 200s is returned.
        """
        response = requests.get(self._url + url, headers=self._headers)
        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            if e.response is not None:
                raise OMMXDA4AdapterError(e.response.text) from e
            else:
                raise OMMXDA4AdapterError(str(e)) from e
        return response.json()

    def post(
        self, url: str, serialized_body: str, headers: dict[str, str]
    ) -> dict[str, Any]:
        """wrapper of `requests.post`.

        :param url: The endpoint path
        :param serialized_body: JSON string to send as the request body
        :param headers: HTTP request headers
        :return: json dict
        :raises OMMXDA4AdapterError: If an HTTP status other than 200s is returned
        """
        response = requests.post(self._url + url, headers=headers, data=serialized_body)
        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            if e.response is not None:
                raise OMMXDA4AdapterError(e.response.text) from e
            else:
                raise OMMXDA4AdapterError(str(e)) from e
        return response.json()

    def delete(self, url: str) -> dict[str, Any]:
        """wrapper of `requests.delete`.

        :param url: The endpoint path
        :return: json dict
        :raises OMMXDA4AdapterError: If an HTTP status other than 200s is returned.
        """
        response = requests.delete(self._url + url, headers=self._headers)
        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            if e.response is not None:
                raise OMMXDA4AdapterError(e.response.text) from e
            else:
                raise OMMXDA4AdapterError(str(e)) from e
        return response.json()

    def post_qubo_solve(
        self,
        qubo_request: QuboRequest,
        *,
        blob_sas_token: Optional[str] = None,
        blob_account_name: Optional[str] = None,
    ) -> str:
        """Post QUBO problem to the DA4 solver with optional Azure Blob Storage support.

        :param qubo_request: The QUBO problem data
        :param blob_sas_token: X-Blob-Sas-Token. Defaults to None.
        :param blob_account_name: X-Storage-Account-Name. Defaults to None.
        :return: The job ID
        """
        serialized_body = qubo_request.model_dump_json(exclude_none=True, by_alias=True)

        if (blob_sas_token is not None) and (blob_account_name is not None):
            da4_headers = {
                "accept": "application/json",
                "Content-Type": "application/json",
                "X-Api-Key": self._token,
                "X-Blob-Sas-Token": blob_sas_token,
                "X-Storage-Account-Name": blob_account_name,
            }
        else:
            da4_headers = {
                "accept": "application/json",
                "Content-Type": "application/json",
                "X-Api-Key": self._token,
            }

        api_url = f"/{self._version}/async/qubo/solve"

        response = self.post(api_url, serialized_body, da4_headers)
        return response["job_id"]

    def get_jobs(self) -> list[dict[str, str]]:
        """Get a list of registered jobs for DA4.

        :return: A list of registered jobs.
        Example response::

            [
                {
                    "job_id": "contract-a001-1234-5678-1a2b3c4d5e6f-123456789012345",
                    "job_status": "Done",
                    "start_time": "2018-07-10T06:02:22Z"
                },
                {
                    "job_id": "contract-a001-1234-5678-1a2b3c4d5e6f-234567890123456",
                    "job_status": "Running",
                    "start_time": "2018-07-10T06:02:24Z"
                },
            ]
        """
        response = self.get(self._api_url_jobs)
        return response["job_status_list"]

    def get_job_result(self, job_id: str) -> dict[str, Any]:
        """Get job results from DA4.

        :param job_id: The job ID
        :return: The job result data
        """
        api_url = self._api_url_jobs + f"/result/{job_id}"
        return self.get(api_url)

    def delete_job_result(self, job_id: str) -> bool:
        """Delete job in DA4.

        :param job_id: The job ID
        :return: True if deletion was successful, False otherwise
        :raises OMMXDA4AdapterError: If an HTTP status other than 200s is returned
        """
        api_url = self._api_url_jobs + f"/result/{job_id}"
        try:
            self.delete(api_url)
            return True
        except OMMXDA4AdapterError:
            return False

    def post_job_cancel(self, job_id: str) -> bool:
        """Cancel job in DA4.

        :param job_id: The job ID
        :return: True if cancellation was successful, False otherwise
        :raises OMMXDA4AdapterError: If an HTTP status other than 200s is returned
        """
        api_url = self._api_url_jobs + "/cancel"
        serialized_body = JobID(job_id=job_id).model_dump_json()
        try:
            self.post(api_url, serialized_body, self._headers)
            return True
        except OMMXDA4AdapterError:
            return False

    def delete_all_jobs(self) -> None:
        """Delete all jobs in DA4."""
        job_list = self.get_jobs()
        for job in job_list:
            self.delete_job_result(job["job_id"])
        return None

    def cancel_all_jobs(self) -> None:
        """Cancel all jobs in DA4."""
        job_list = self.get_jobs()
        for job in job_list:
            self.post_job_cancel(job["job_id"])
        return None

    def fetch_job_result(self, job_id: str) -> QuboResponse:
        """Fetch job result while waiting for completion.

        This method polls the job status until completion and returns the result.
        When completed, it automatically deletes the job.

        :arg job_id: The job id.
        :return: Result returned from DA4.
        :raises OMMXDA4AdapterError: Raised when solving fails for some reason on DA4.

        """
        for _ in range(self._max_loop_count):
            job_result = self.get_job_result(job_id)
            job_status = job_result["status"]

            if job_status == "Done":
                self.delete_job_result(job_id)
                return QuboResponse(**job_result)
            elif job_status == "Error":
                raise OMMXDA4AdapterError(f"Failed to solve: `{job_result}`.")
            elif (job_status == "Running") or (job_status == "Waiting"):
                time.sleep(self._solver_access_interval)
            else:
                raise OMMXDA4AdapterError(
                    f"Unknown error occurred. The status `{job_status}` is unknown."
                )

        delete_success = self.delete_job_result(job_id)
        cancel_success = self.post_job_cancel(job_id)

        if delete_success:
            raise OMMXDA4AdapterError(
                f"Too long solving time. Job id `{job_id}` has been deleted. Sorry."
            )
        elif cancel_success:
            raise OMMXDA4AdapterError(
                f"Too long solving time. Job id `{job_id}` has been canceled. Sorry."
            )
        else:
            raise OMMXDA4AdapterError(
                f"Too long solving time. Job id `{job_id}` is still running."
            )

    def sample(
        self,
        qubo_request: QuboRequest,
        *,
        blob_sas_token: Optional[str] = None,
        blob_account_name: Optional[str] = None,
    ) -> QuboResponse:
        """Sample the result in DA4.

        :param qubo_request: The QUBO problem data
        :param blob_sas_token: X-Blob-Sas-Token. Defaults to None.
        :param blob_account_name: X-Storage-Account-Name. Defaults to None.
        :return: Result returned from DA4.
        """
        posted_job_id = self.post_qubo_solve(
            qubo_request=qubo_request,
            blob_sas_token=blob_sas_token,
            blob_account_name=blob_account_name,
        )

        return self.fetch_job_result(posted_job_id)



from fastapi import Header, status
from fastapi.responses import JSONResponse

from .models import Accept, ContentType


MOCK_API_KEY = "mock-api-key"


def check_auth(
    access_token: str = Header(None, alias="X-Access-Token"),
    api_key: str = Header(None, alias="X-Api-Key"),
) -> JSONResponse | None:
    if access_token is None and api_key is None:
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content={
                "error": {
                    "code": "400",
                    "title": "Bad Request",
                    "message": "Neither X-Access-Token nor X",
                }
            },
        )

    if access_token is not None and api_key is not None:
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content={
                "error": {
                    "code": "400",
                    "title": "Bad Request",
                    "message": "Choose either X-Access-Token or X-Api-Key",
                }
            },
        )

    # NOTE: This mock server only authorizes `"X-Api-Key": MOCK_API_KEY`
    if api_key is None or api_key != MOCK_API_KEY:
        return JSONResponse(
            status_code=status.HTTP_401_UNAUTHORIZED,
            content={
                "error": {
                    "code": "401",
                    "title": "Unauthorized",
                    "message": "Access denied due to invalid header",
                }
            },
        )

    return None


def check_accept(
    accept: Accept = Header(None, alias="Accept"),
) -> JSONResponse | None:
    """
    NOTE:
    This check does not exist in DA4, but was added because we want to make accept: application/json mandatory.
    """
    if accept is None:
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content={
                "error": {
                    "code": "400",
                    "title": "Bad Request",
                    "message": "Accept header is required",
                }
            },
        )

    if accept != Accept.json:
        return JSONResponse(
            status_code=status.HTTP_406_NOT_ACCEPTABLE,
            content={
                "error": {
                    "code": "406",
                    "title": "Not Acceptable",
                    "message": "Accept header is not application/json",
                }
            },
        )

    return None


def check_content_type(
    content_type: ContentType = Header(None, alias="Content-Type"),
) -> JSONResponse | None:
    """
    NOTE:
    This check does not exist in DA4, but was added because we want to make Content-Type: application/json mandatory.
    """
    if content_type is None:
        return JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST,
            content={
                "error": {
                    "code": "400",
                    "title": "Bad Request",
                    "message": "Content-Type header is required",
                }
            },
        )

    if content_type != ContentType.json:
        return JSONResponse(
            status_code=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
            content={
                "error": {
                    "code": "415",
                    "title": "Unsupported Media Type",
                    "message": "Content-Type header is not application/json",
                }
            },
        )

    return None



from typing import List, Literal, Optional, Union

from ommx.adapter import SamplerAdapter
from ommx.v1 import (
    Instance,
    SampleSet,
    Constraint,
    DecisionVariable,
    State,
    Samples,
)

from .client import DA4Client
from .exception import OMMXDA4AdapterError
from .models import (
    BinaryPolynomial,
    BinaryPolynomialTerm,
    FujitsuDA3Solver,
    Inequalities,
    PenaltyBinaryPolynomial,
    QuboRequest,
    QuboResponse,
)


class OMMXDA4Adapter(SamplerAdapter):
    def __init__(
        self,
        ommx_instance: Instance,
        *,
        time_limit_sec: int = 10,
        target_energy: Optional[float] = None,
        num_run: int = 16,
        num_group: int = 1,
        num_output_solution: int = 5,
        gs_level: int = 5,
        gs_cutoff: int = 8000,
        one_hot_level: int = 3,
        one_hot_cutoff: int = 100,
        penalty_auto_mode: int = 1,
        penalty_coef: int = 1,
        penalty_inc_rate: int = 150,
        max_penalty_coef: int = 0,
        guidance_config: Optional[dict[str, bool]] = None,
        fixed_config: Optional[dict[str, bool]] = None,
        inequalities_lambda: Optional[dict[int, int]] = None,
    ):
        """Digital Annealer adapter for OMMX.

        :param ommx_instance: OMMX instance
        :param time_limit_sec: Upper limit of execution time (in seconds)
        :param target_energy: Target energy to be terminated when reached
        :param num_run: Number of parallel trial iterations
        :param num_group: Number of groups of parallel trials
        :param num_output_solution: Number of output solutions for each parallel trial group
        :param gs_level: Levels of global search
        :param gs_cutoff: Convergence decision level for global search. If 0 is set, this function is turned off
        :param one_hot_level: Levels of 1hot constraints search
        :param one_hot_cutoff: Convergence decision level for 1hot constraints search. If 0 is set, this function is turned off
        :param penalty_auto_mode: Coefficient adjustment mode for constraint terms. If set to 0, no adjustment is made
        :param penalty_coef: Coefficient of the constraint term
        :param penalty_inc_rate: Parameters for automatic adjustment of constraint terms
        :param max_penalty_coef: Maximum value of constraint term coefficients. If 0 is set, there is no maximum value
        :param guidance_config: Initial value of each variable
        :param fixed_config: Fixed value for each variable
        :param inequalities_lambda: Coefficient of inequality. If omitted, set to 1. Defaults to None.
        """
        self._ommx_instance = ommx_instance
        self._inequalities_lambda = inequalities_lambda

        self._check_decision_variable()
        self._one_hot_dict = self._generate_one_hot_dict()
        self._variable_map = self._generate_variable_map()

        # generate QuboRequest from OMMX instance
        self._sampler_input = QuboRequest(
            fujitsuDA3=FujitsuDA3Solver(
                time_limit_sec=time_limit_sec,
                target_energy=target_energy,
                num_run=num_run,
                num_group=num_group,
                num_output_solution=num_output_solution,
                gs_level=gs_level,
                gs_cutoff=gs_cutoff,
                one_hot_level=one_hot_level,
                one_hot_cutoff=one_hot_cutoff,
                internal_penalty=self._generate_internal_penalty(),
                penalty_auto_mode=penalty_auto_mode,
                penalty_coef=penalty_coef,
                penalty_inc_rate=penalty_inc_rate,
                max_penalty_coef=max_penalty_coef,
                guidance_config=guidance_config,
                fixed_config=fixed_config,
                one_way_one_hot_groups=self._generate_one_way_one_hot_groups(),
            ),
            binary_polynomial=self._generate_binary_polynomial(),
            penalty_binary_polynomial=self._generate_penalty_binary_polynomial(),
            inequalities=self._generate_inequalities(),
        )

    @property
    def sampler_input(self) -> QuboRequest:
        """Get QuboRequest from OMMX instance.

        :return: QuboRequest
        """
        return self._sampler_input

    @classmethod
    def sample(
        cls,
        ommx_instance: Instance,
        *,
        token: Optional[str] = None,
        url: str = "https://api.aispf.global.fujitsu.com/da",
        version: Literal["v4", "v3c"] = "v4",
    ) -> SampleSet:
        """Sample the result in DA4 with DA4Client.

        :param ommx_instance: OMMX instance
        :param token: Authentication token for DA4 API. Defaults to None.
        :param url: URL to the Fujitsu Digital Annealer. Defaults to "https://api.aispf.global.fujitsu.com/da".
        :param version: The version of Digital Annealer as either "v4" or "v3c". Defaults to "v4".
        :return: SampleSet
        """
        if token is None:
            raise OMMXDA4AdapterError(
                "token is required. Please set the token to use the DA4 API."
            )

        adapter = cls(ommx_instance)
        qubo_request = adapter.sampler_input
        client = DA4Client(token=token, url=url, version=version)
        qubo_response = client.sample(qubo_request=qubo_request)

        return adapter.decode_to_sampleset(qubo_response)

    def decode_to_sampleset(self, data: QuboResponse) -> SampleSet:
        """Decode QuboResponse to SampleSet.

        :param data: The QUBO result data from DA4
        :return: SampleSet
        """

        sample_id: int = 0
        entries: List[Samples.SamplesEntry] = []

        reversed_variable_map = {v: k for k, v in self._variable_map.items()}

        for solution in data.qubo_solution.solutions:
            configuration = solution.configuration
            try:
                converted_configuration = {
                    reversed_variable_map[int(k)]: int(v)
                    for k, v in configuration.items()
                }
            except KeyError as e:
                raise OMMXDA4AdapterError(
                    f"Invalid solution configuration: The solution contains an unexpected decision variable id ({e})."
                )
            next_sample_id = sample_id + solution.frequency
            entries.append(
                Samples.SamplesEntry(
                    state=State(entries=converted_configuration),
                    ids=[i for i in range(sample_id, next_sample_id)],
                )
            )
            sample_id = next_sample_id

        samples = Samples(
            entries=entries,
        )

        return self._ommx_instance.evaluate_samples(samples)

    def _check_decision_variable(self):
        """Check if the decision variables are binary."""
        instance = self._ommx_instance

        for decision_variable in instance.get_decision_variables():
            if decision_variable.kind != DecisionVariable.BINARY:
                raise OMMXDA4AdapterError(
                    f"The decision variable must be binary: id {decision_variable.id}"
                )

    def _generate_binary_polynomial(self) -> BinaryPolynomial:
        """Generate BinaryPolynomial from OMMX instance."

        :return: BinaryPolynomial
        """
        instance = self._ommx_instance

        function = instance.objective

        # if sense is maximize, multiply by -1 (DA4 only supports minimization)
        if instance.sense == Instance.MAXIMIZE:
            function = -instance.objective

        # get objective terms
        terms = function.terms

        binary_polynomial_terms = [
            BinaryPolynomialTerm(
                c=value, p=self._replace_polynomials_with_variable_map(key)
            )
            for key, value in terms.items()
        ]

        return BinaryPolynomial(terms=binary_polynomial_terms)

    def _generate_penalty_binary_polynomial(
        self,
    ) -> Union[PenaltyBinaryPolynomial, None]:
        """Generate PenaltyBinaryPolynomial from OMMX instance.

        Example:
        =========
        Original: x₀ + x₁ - 1
        Squared: (x₀ + x₁ - 1)² = x₀² + 2x₀x₁ - 2x₀ + x₁² - 2x₁ + 1
        After binary simplification: x₀ + 2x₀x₁ - 2x₀ + x₁ - 2x₁ + 1 = 2x₀x₁ - x₀ - x₁ + 1
        Dictionary form: {(0, 1): 2.0, (0,): -1.0, (1,): -1.0, (): 1.0}

        :return: PenaltyBinaryPolynomial
        """
        instance = self._ommx_instance

        # Squared Polynomial with Binary Variables
        squared_terms_dict = {}
        for constraint in instance.get_constraints():
            # skip if not equality constraints
            if constraint.equality != Constraint.EQUAL_TO_ZERO:
                continue

            if constraint.id in self._one_hot_dict:
                continue

            function = constraint.function
            squared_function = function * function

            for key, value in squared_function.terms.items():
                # Remove duplicates by converting to set
                set_key = set(key)
                list_key = tuple(set_key)

                # Add to existing value when duplicates occur
                if list_key in squared_terms_dict:
                    squared_terms_dict[list_key] += value
                else:
                    squared_terms_dict[list_key] = value

        penalty_binary_polynomial_terms = [
            BinaryPolynomialTerm(
                c=value, p=self._replace_polynomials_with_variable_map(key)
            )
            for key, value in squared_terms_dict.items()
        ]

        if len(penalty_binary_polynomial_terms) == 0:
            return None
        else:
            return PenaltyBinaryPolynomial(terms=penalty_binary_polynomial_terms)

    def _generate_inequalities(self) -> Union[list[Inequalities], None]:
        """Generate Inequalities from OMMX instance.

        :return: Inequalities
        """
        instance = self._ommx_instance

        inequalities_list = []
        for constraint in instance.get_constraints():
            # skip if not inequality constraints
            if constraint.equality != Constraint.LESS_THAN_OR_EQUAL_TO_ZERO:
                continue

            terms = constraint.function.terms
            inequalities_terms = [
                BinaryPolynomialTerm(
                    c=value, p=self._replace_polynomials_with_variable_map(key)
                )
                for key, value in terms.items()
            ]

            if (
                self._inequalities_lambda is None
                or constraint.id not in self._inequalities_lambda
            ):
                lambda_ = 1
            else:
                lambda_ = self._inequalities_lambda[constraint.id]
            inequalities_list.append(
                Inequalities(terms=inequalities_terms, **{"lambda": lambda_})
            )

        if len(inequalities_list) == 0:
            return None
        else:
            return inequalities_list

    def _generate_internal_penalty(self) -> int:
        """Generate internal penalty

        if set one way one hot or two way one hot, internal_penalty is 1
        else, internal_penalty is 0

        caution: two way one hot is not supported yet
        """
        instance = self._ommx_instance

        if len(instance.raw.constraint_hints.one_hot_constraints) == 0:
            return 0
        else:
            return 1

    def _generate_variable_map(self) -> dict[int, int]:
        """Generate variable map that represents the correspondence
        between the IDs of decision variables in ommx.v1.Instance and the variable numbers on QuboRequest.
        """
        instance = self._ommx_instance
        variable_map = {}

        # First enumerate the decision variables from one-hot constraints,
        # then enumerate the remaining decision variables afterwards.
        index = 0

        for variables in self._one_hot_dict.values():
            for variable in variables:
                variable_map[variable] = index
                index += 1
        for decision_variable in instance.get_decision_variables():
            # skip if already in variable_map
            if decision_variable.id in variable_map:
                continue
            variable_map[decision_variable.id] = index
            index += 1

        return variable_map

    def _replace_polynomials_with_variable_map(
        self, polynomial: tuple[int, ...]
    ) -> list[int]:
        """Replace the IDs of decision variables in ommx.v1.Instance
        with variable numbers on QuboRequest using variable map.

        variable map is generated by _generate_variable_map().
        This corresponds between the IDs of decision variables in ommx.v1.Instance
        and the variable numbers on QuboRequest.
        """
        transformed_polynomial = [self._variable_map[p] for p in polynomial]
        return transformed_polynomial

    def _generate_one_way_one_hot_groups(
        self,
    ) -> Union[dict[Literal["numbers"], list[int]], None]:
        """Generate one way one hot groups."""

        numbers = [len(variables) for variables in self._one_hot_dict.values()]

        if len(numbers) == 0:
            return None
        else:
            return {"numbers": numbers}

    def _generate_one_hot_dict(self) -> dict[int, list[int]]:
        """Generate a dictionary of one-hot constraints without duplicate decision variables.

        Examples:
        =========
        case 1：no duplicate decision variables
        constraint_1: id=0, x₀ + x₁ + x₂ = 1
        constraint_2: id=1, x₃ + x₄ = 1
        constraint_3: id=2, x₅ + x₆ = 1
        one_hot_dict = {
                            0: [0, 1, 2],
                            1: [3, 4],
                            2: [5, 6],
                        }

        case 2：duplicate decision variables
        Prioritize longer constraints and exclude shorter ones
        constraint_1: id=0, x₀ + x₁ + x₂ = 1
        constraint_2: id=1, x₁ + x₃ = 1
        one_hot_dict = {
                            0: [0, 1, 2]
                        }
        """
        instance = self._ommx_instance

        sorted_one_hot_constraints = sorted(
            instance.raw.constraint_hints.one_hot_constraints,
            key=lambda x: len(x.decision_variables),
            reverse=True,
        )

        one_hot_dict: dict[int, list[int]] = {}
        used_variables = set()
        for one_hot_constraint in sorted_one_hot_constraints:
            if any(
                [
                    variable in used_variables
                    for variable in one_hot_constraint.decision_variables
                ]
            ):
                continue

            used_variables.update(one_hot_constraint.decision_variables)
            one_hot_dict[one_hot_constraint.constraint_id] = [
                variable_id for variable_id in one_hot_constraint.decision_variables
            ]

        return one_hot_dict



import asyncio
import datetime
from typing import Literal

from .models import QuboRequest, QuboResponse
from .random_qubo_response import random_qubo_response


class Job:
    def __init__(
        self,
        job_id: str,
        qubo_request: QuboRequest,
        start_time: str,
    ):
        self._job_id = job_id
        self._qubo_request = qubo_request
        self._start_time = start_time

    @property
    def status(self) -> Literal["Waiting", "Running", "Done", "Canceled", "Error"]:
        raise NotImplementedError()

    def get_job_status_info(self) -> dict[str, str]:
        return {
            "job_id": self._job_id,
            "job_status": self.status,
            "start_time": self._start_time,
        }


class WaitingJob(Job):
    @property
    def status(self) -> Literal["Waiting"]:
        return "Waiting"

    def run(self) -> "RunningJob":
        return RunningJob(
            job_id=self._job_id,
            qubo_request=self._qubo_request,
            start_time=self._start_time,
        )

    def cancel(self) -> "CanceledJob":
        return CanceledJob(
            job_id=self._job_id,
            qubo_request=self._qubo_request,
            start_time=self._start_time,
        )


class RunningJob(Job):
    @property
    def status(self) -> Literal["Running"]:
        return "Running"

    def done(self) -> "DoneJob":
        return DoneJob(
            job_id=self._job_id,
            qubo_request=self._qubo_request,
            start_time=self._start_time,
        )

    def error(self) -> "ErrorJob":
        return ErrorJob(
            job_id=self._job_id,
            qubo_request=self._qubo_request,
            start_time=self._start_time,
        )


class DoneJob(Job):
    @property
    def status(self) -> Literal["Done"]:
        return "Done"

    def get_result(self) -> QuboResponse:
        return random_qubo_response(qubo_request=self._qubo_request)


class CanceledJob(Job):
    @property
    def status(self) -> Literal["Canceled"]:
        return "Canceled"


class ErrorJob(Job):
    @property
    def status(self) -> Literal["Error"]:
        return "Error"


async def job_runner(
    job_id: str,
    qubo_request: QuboRequest,
    job_store: dict[str, Job],
) -> None:
    WAITING_TIME = 5
    RUNNING_TIME = 10

    waiting_job = WaitingJob(
        job_id=job_id,
        qubo_request=qubo_request,
        start_time=datetime.datetime.now().isoformat(),
    )
    job_store[job_id] = waiting_job
    print(f"[CREATE] job_id: {job_id}, status: {waiting_job.status}")
    await asyncio.sleep(WAITING_TIME)

    if isinstance(job_store[job_id], CanceledJob):
        return

    running_job = waiting_job.run()
    job_store[job_id] = running_job
    print(f"[UPDATE] job_id: {job_id}, status: {running_job.status}")
    await asyncio.sleep(RUNNING_TIME)

    if isinstance(job_store[job_id], CanceledJob):
        return

    done_job = running_job.done()
    job_store[job_id] = done_job
    print(f"[UPDATE] job_id: {job_id}, status: {done_job.status}")



from enum import Enum
from typing import Literal

import pydantic


class ContentType(str, Enum):
    json = "application/json"


class Accept(str, Enum):
    json = "application/json"


class FujitsuDA3Solver(pydantic.BaseModel):
    time_limit_sec: int = 10
    target_energy: float | None = None
    num_run: int = 16
    num_group: int = 1
    num_output_solution: int = 5
    gs_level: int = 5
    gs_cutoff: int = 8000
    one_hot_level: int = 3
    one_hot_cutoff: int = 100
    internal_penalty: int = 0
    penalty_auto_mode: int = 1
    penalty_coef: int = 1
    penalty_inc_rate: int = 150
    max_penalty_coef: int = 0
    guidance_config: dict[str, bool] = {}
    fixed_config: dict[str, bool] = {}
    one_way_one_hot_groups: dict[Literal["numbers"], list[int]] = {}
    two_way_one_hot_groups: dict[Literal["numbers"], list[int]] = {}

    @pydantic.field_validator("time_limit_sec")
    @classmethod
    def check_time_limit_sec(cls, v):
        if not (1 <= v <= 3600):
            raise ValueError("`time_limit_sec` must be between 1 and 3600")
        return v

    @pydantic.field_validator("target_energy")
    @classmethod
    def check_target_energy(cls, v):
        if v is None:
            return v

        # NOTE: Limited to a range where no numerical errors occur in DA4.
        if not (-pow(2, 52) <= v <= pow(2, 52)):
            raise ValueError("`target_energy` must be between - 2^52 and 2^52")
        return v

    @pydantic.field_validator("num_run")
    @classmethod
    def check_num_run(cls, v):
        if not (1 <= v <= 1024):
            raise ValueError("`num_run` must be between 1 and 1024")
        return v

    @pydantic.field_validator("num_group")
    @classmethod
    def check_num_group(cls, v):
        if not (1 <= v <= 16):
            raise ValueError("`num_group` must be between 1 and 16")
        return v

    @pydantic.field_validator("num_output_solution")
    @classmethod
    def check_num_output_solution(cls, v):
        if not (1 <= v <= 1024):
            raise ValueError("`num_output_solution` must be between 1 and 1024")
        return v

    @pydantic.field_validator("gs_level")
    @classmethod
    def check_gs_level(cls, v):
        if not (0 <= v <= 100):
            raise ValueError("`gs_level` must be between 0 and 100")
        return v

    @pydantic.field_validator("gs_cutoff")
    @classmethod
    def check_gs_cutoff(cls, v):
        if not (0 <= v <= 1000000):
            raise ValueError("`gs_cutoff` must be between 0 and 1000000")
        return v

    @pydantic.field_validator("one_hot_level")
    @classmethod
    def check_one_hot_level(cls, v):
        if not (0 <= v <= 100):
            raise ValueError("`one_hot_level` must be between 0 and 100")
        return v

    @pydantic.field_validator("one_hot_cutoff")
    @classmethod
    def check_one_hot_cutoff(cls, v):
        if not (0 <= v <= 1000000):
            raise ValueError("`one_hot_cutoff` must be between 0 and 1000000")
        return v

    @pydantic.field_validator("internal_penalty")
    @classmethod
    def check_internal_penalty(cls, v):
        if not (0 <= v <= 1):
            raise ValueError("`internal_penalty` must be 0 or 1")
        return v

    @pydantic.field_validator("penalty_auto_mode")
    @classmethod
    def check_penalty_auto_mode(cls, v):
        if not (0 <= v <= 10000):
            raise ValueError("`penalty_auto_mode` must be between 0 and 10000")
        return v

    @pydantic.field_validator("penalty_coef")
    @classmethod
    def check_penalty_coef(cls, v):
        if not (1 <= v <= 9223372036854775807):
            raise ValueError("`penalty_coef` must be between 1 and 9223372036854775807")
        return v

    @pydantic.field_validator("penalty_inc_rate")
    @classmethod
    def check_penalty_inc_rate(cls, v):
        if not (100 <= v <= 200):
            raise ValueError("`penalty_inc_rate` must be between 100 and 200")
        return v

    @pydantic.field_validator("max_penalty_coef")
    @classmethod
    def check_max_penalty_coef(cls, v):
        if not (0 <= v <= 9223372036854775807):
            raise ValueError(
                "`max_penalty_coef` must be between 0 and 9223372036854775807"
            )
        return v


class BinaryPolynomialTerm(pydantic.BaseModel):
    """
    NOTE:
    In DA4, `coefficient` and `polynomials` can be used instead of `c` and `p`.
    However, since `c` and `p` are used with priority, we limit the attributes to `c` and `p` here.
    """

    c: float
    p: list[int]

    @pydantic.field_validator("c")
    @classmethod
    def check_c(cls, v):
        # NOTE: Limited to a range where no numerical errors occur in DA4.
        if not (-pow(2, 52) <= v <= pow(2, 52)):
            raise ValueError("`c` must be between - 2^52 and 2^52")
        return v


class BinaryPolynomial(pydantic.BaseModel):
    terms: list[BinaryPolynomialTerm]


class PenaltyBinaryPolynomial(pydantic.BaseModel):
    terms: list[BinaryPolynomialTerm]


class Inequality(pydantic.BaseModel):
    terms: list[BinaryPolynomialTerm]
    lambda_: int = pydantic.Field(alias="lambda", default=1)

    @pydantic.field_validator("lambda_")
    @classmethod
    def check_lambda_(cls, v):
        if not (1 <= v <= 1000000000):
            raise ValueError("`lambda` must be between 1 and 1000000000 ")
        return v


class QuboRequest(pydantic.BaseModel):
    fujitsuDA3: FujitsuDA3Solver
    binary_polynomial: BinaryPolynomial | None = None
    penalty_binary_polynomial: PenaltyBinaryPolynomial | None = None
    inequalities: list[Inequality] | None = None
    object_format: str | None = None
    bucket_name: str | None = None
    binary_polynomial_object_name: str | None = None
    penalty_binary_polynomial_object_name: str | None = None
    inequalities_object_name: str | None = None


class Progress(pydantic.BaseModel):
    energy: float
    penalty_energy: float
    time: float


class QuboSolution(pydantic.BaseModel):
    energy: float
    penalty_energy: float
    frequency: int
    configuration: dict[str, bool]


class SolverTiming(pydantic.BaseModel):
    solve_time: str
    total_elapsed_time: str


class QuboSolutionList(pydantic.BaseModel):
    progress: list[Progress]
    result_status: bool
    solutions: list[QuboSolution]
    timing: SolverTiming


class QuboResponse(pydantic.BaseModel):
    qubo_solution: QuboSolutionList
    status: Literal["Done", "Deleted"]


class QuboErrorResponse(pydantic.BaseModel):
    message: str
    status: Literal["Done", "Deleted"]


class JobID(pydantic.BaseModel):
    job_id: str


class JobStatus(pydantic.BaseModel):
    status: Literal["Waiting", "Running", "Done", "Canceled", "Error"]


class JobStatusInfo(pydantic.BaseModel):
    job_id: str
    job_status: Literal["Waiting", "Running", "Done", "Canceled", "Error"]
    start_time: str


class JobStatusList(pydantic.BaseModel):
    job_status_list: list[JobStatusInfo]



import uuid

from fastapi import BackgroundTasks, Depends, FastAPI, status
from fastapi.responses import JSONResponse

from .check import check_auth, check_accept, check_content_type
from .job import Job, job_runner, DoneJob, WaitingJob
from .models import QuboRequest, JobStatusList, JobID, JobStatus


app = FastAPI()
job_store: dict[str, Job] = {}


@app.get("/v1/healthcheck", tags=["v1"])
def get_v1_healthcheck(
    auth: JSONResponse | None = Depends(check_auth),
    accept: JSONResponse | None = Depends(check_accept),
) -> JSONResponse:
    if auth is not None:
        return auth
    if accept is not None:
        return accept

    return JSONResponse(
        status_code=status.HTTP_200_OK,
        content={},
    )


@app.post("/v4/async/qubo/solve", tags=["v4"])
def post_v4_async_qubo_solve(
    qubo_request: QuboRequest,
    background_tasks: BackgroundTasks,
    auth: JSONResponse | None = Depends(check_auth),
    accept: JSONResponse | None = Depends(check_accept),
    content_type: JSONResponse | None = Depends(check_content_type),
) -> JSONResponse:
    if auth is not None:
        return auth
    if accept is not None:
        return accept
    if content_type is not None:
        return content_type

    # TODO: Write a process to return an error when the number of jobs exceeds a certain limit

    job_id = uuid.uuid4().hex
    background_tasks.add_task(job_runner, job_id, qubo_request, job_store)

    return JSONResponse(
        status_code=status.HTTP_200_OK,
        content=JobID(job_id=job_id).model_dump(),
    )


@app.get("/v4/async/jobs", tags=["v4"])
def get_v4_async_jobs(
    auth: JSONResponse | None = Depends(check_auth),
    accept: JSONResponse | None = Depends(check_accept),
) -> JSONResponse:
    if auth is not None:
        return auth
    if accept is not None:
        return accept

    job_status_list = []
    for _, value in job_store.items():
        job_status_list.append(value.get_job_status_info())

    return JSONResponse(
        status_code=status.HTTP_200_OK,
        content=JobStatusList(job_status_list=job_status_list).model_dump(),
    )


@app.get("/v4/async/jobs/result/{job_id}", tags=["v4"])
def get_v4_async_jobs_result(
    job_id: str,
    auth: JSONResponse | None = Depends(check_auth),
    accept: JSONResponse | None = Depends(check_accept),
) -> JSONResponse:
    if auth is not None:
        return auth
    if accept is not None:
        return accept

    if job_id not in job_store:
        return JSONResponse(
            status_code=status.HTTP_404_NOT_FOUND,
            content={
                "error": {
                    "code": 404,
                    "title": "Resource Not Found",
                    "message": "Resource not found.",
                }
            },
        )

    job = job_store[job_id]
    if isinstance(job, DoneJob):
        return JSONResponse(
            status_code=status.HTTP_200_OK,
            content=job.get_result().model_dump(),
        )
    else:
        return JSONResponse(
            status_code=status.HTTP_200_OK,
            content=JobStatus(status=job.status).model_dump(),
        )


@app.delete("/v4/async/jobs/result/{job_id}", tags=["v4"])
def delete_v4_async_jobs_result(
    job_id: str,
    auth: JSONResponse | None = Depends(check_auth),
    accept: JSONResponse | None = Depends(check_accept),
) -> JSONResponse:
    if auth is not None:
        return auth
    if accept is not None:
        return accept

    if job_id not in job_store:
        return JSONResponse(
            status_code=status.HTTP_404_NOT_FOUND,
            content={
                "error": {
                    "code": 404,
                    "title": "Resource Not Found",
                    "message": "Resource not found.",
                }
            },
        )

    job = job_store.pop(job_id)
    if isinstance(job, DoneJob):
        qubo_response = job.get_result()
        qubo_response.status = "Deleted"
        return JSONResponse(
            status_code=status.HTTP_200_OK,
            content=qubo_response.model_dump(),
        )
    else:
        return JSONResponse(
            status_code=status.HTTP_200_OK,
            content=JobStatus(status=job.status).model_dump(),
        )


@app.post("/v4/async/jobs/cancel", tags=["v4"])
def post_v4_async_jobs_cancel(
    cancel_request: JobID,
    auth: JSONResponse | None = Depends(check_auth),
    accept: JSONResponse | None = Depends(check_accept),
) -> JSONResponse:
    if auth is not None:
        return auth
    if accept is not None:
        return accept

    job_id = cancel_request.job_id
    if job_id not in job_store:
        return JSONResponse(
            status_code=status.HTTP_404_NOT_FOUND,
            content={
                "error": {
                    "code": 404,
                    "title": "Resource Not Found",
                    "message": "Resource not found.",
                }
            },
        )

    job = job_store[job_id]
    if isinstance(job, WaitingJob):
        job_store[job_id] = job.cancel()
        return JSONResponse(
            status_code=status.HTTP_200_OK,
            content=JobStatus(status=job_store[job_id].status).model_dump(),
        )
    else:
        return JSONResponse(
            status_code=status.HTTP_200_OK,
            content=JobStatus(status=job.status).model_dump(),
        )



from typing import Literal, Optional, List, Dict

import pydantic


class FujitsuDA3Solver(pydantic.BaseModel):
    time_limit_sec: int = 10
    target_energy: Optional[float] = None
    num_run: int = 16
    num_group: int = 1
    num_output_solution: int = 5
    gs_level: int = 5
    gs_cutoff: int = 8000
    one_hot_level: int = 3
    one_hot_cutoff: int = 100
    internal_penalty: int = 0
    penalty_auto_mode: int = 1
    penalty_coef: int = 1
    penalty_inc_rate: int = 150
    max_penalty_coef: int = 0
    guidance_config: Optional[Dict[str, bool]] = None
    fixed_config: Optional[Dict[str, bool]] = None
    one_way_one_hot_groups: Optional[Dict[Literal["numbers"], List[int]]] = None
    two_way_one_hot_groups: Optional[Dict[Literal["numbers"], List[int]]] = None

    @pydantic.field_validator("time_limit_sec")
    @classmethod
    def check_time_limit_sec(cls, v):
        if not (1 <= v <= 3600):
            raise ValueError("`time_limit_sec` must be between 1 and 3600")
        return v

    @pydantic.field_validator("target_energy")
    @classmethod
    def check_target_energy(cls, v):
        if v is None:
            return v

        if not (-pow(2, 126) <= v <= pow(2, 126)):
            raise ValueError("`target_energy` must be between - 2^126 and 2^126")
        return v

    @pydantic.field_validator("num_run")
    @classmethod
    def check_num_run(cls, v):
        if not (1 <= v <= 1024):
            raise ValueError("`num_run` must be between 1 and 1024")
        return v

    @pydantic.field_validator("num_group")
    @classmethod
    def check_num_group(cls, v):
        if not (1 <= v <= 16):
            raise ValueError("`num_group` must be between 1 and 16")
        return v

    @pydantic.field_validator("num_output_solution")
    @classmethod
    def check_num_output_solution(cls, v):
        if not (1 <= v <= 1024):
            raise ValueError("`num_output_solution` must be between 1 and 1024")
        return v

    @pydantic.field_validator("gs_level")
    @classmethod
    def check_gs_level(cls, v):
        if not (0 <= v <= 100):
            raise ValueError("`gs_level` must be between 0 and 100")
        return v

    @pydantic.field_validator("gs_cutoff")
    @classmethod
    def check_gs_cutoff(cls, v):
        if not (0 <= v <= 1000000):
            raise ValueError("`gs_cutoff` must be between 0 and 1000000")
        return v

    @pydantic.field_validator("one_hot_level")
    @classmethod
    def check_one_hot_level(cls, v):
        if not (0 <= v <= 100):
            raise ValueError("`one_hot_level` must be between 0 and 100")
        return v

    @pydantic.field_validator("one_hot_cutoff")
    @classmethod
    def check_one_hot_cutoff(cls, v):
        if not (0 <= v <= 1000000):
            raise ValueError("`one_hot_cutoff` must be between 0 and 1000000")
        return v

    @pydantic.field_validator("internal_penalty")
    @classmethod
    def check_internal_penalty(cls, v):
        if not (0 <= v <= 1):
            raise ValueError("`internal_penalty` must be 0 or 1")
        return v

    @pydantic.field_validator("penalty_auto_mode")
    @classmethod
    def check_penalty_auto_mode(cls, v):
        if not (0 <= v <= 10000):
            raise ValueError("`penalty_auto_mode` must be between 0 and 10000")
        return v

    @pydantic.field_validator("penalty_coef")
    @classmethod
    def check_penalty_coef(cls, v):
        if not (1 <= v <= 9223372036854775807):
            raise ValueError("`penalty_coef` must be between 1 and 9223372036854775807")
        return v

    @pydantic.field_validator("penalty_inc_rate")
    @classmethod
    def check_penalty_inc_rate(cls, v):
        if not (100 <= v <= 200):
            raise ValueError("`penalty_inc_rate` must be between 100 and 200")
        return v

    @pydantic.field_validator("max_penalty_coef")
    @classmethod
    def check_max_penalty_coef(cls, v):
        if not (0 <= v <= 9223372036854775807):
            raise ValueError(
                "`max_penalty_coef` must be between 0 and 9223372036854775807"
            )
        return v


class BinaryPolynomialTerm(pydantic.BaseModel):
    """
    NOTE:
    実際のDAでは、 `c` と `p` の代わりに `cofficient` と `polynomials` を利用できる。
    しかし、 `c` と `p` が優先して利用されるため、ここでは `c` と `p` に属性を限定している。
    """

    c: float
    p: List[int]

    @pydantic.field_validator("c")
    @classmethod
    def check_c(cls, v):
        """
        NOTE: 推奨値の範囲に制限している。実際はもっと広い範囲の値を入れることができる。
        """
        if not (-pow(2, 52) <= v <= pow(2, 52)):
            raise ValueError("`c` must be between - 2^52 and 2^52")
        return v


class BinaryPolynomial(pydantic.BaseModel):
    terms: List[BinaryPolynomialTerm]


class PenaltyBinaryPolynomial(pydantic.BaseModel):
    terms: List[BinaryPolynomialTerm]


class Inequalities(pydantic.BaseModel):
    terms: List[BinaryPolynomialTerm]
    lambda_: int = pydantic.Field(alias="lambda", default=1)

    @pydantic.field_validator("lambda_")
    @classmethod
    def check_lambda_(cls, v):
        if not (1 <= v <= 1000000000):
            raise ValueError("`lambda` must be between 1 and 1000000000 ")
        return v


class QuboRequest(pydantic.BaseModel):
    fujitsuDA3: FujitsuDA3Solver
    binary_polynomial: Optional[BinaryPolynomial] = None
    penalty_binary_polynomial: Optional[PenaltyBinaryPolynomial] = None
    inequalities: Optional[List[Inequalities]] = None
    bucket_name: Optional[str] = None
    binary_polynomial_object_name: Optional[str] = None
    penalty_binary_polynomial_object_name: Optional[str] = None
    inequalities_object_name: Optional[str] = None


class Progress(pydantic.BaseModel):
    energy: float
    penalty_energy: float
    time: float


class QuboSolution(pydantic.BaseModel):
    energy: float
    penalty_energy: float
    frequency: int
    configuration: Dict[str, bool]


class SolverTiming(pydantic.BaseModel):
    solve_time: str
    total_elapsed_time: str


class QuboSolutionList(pydantic.BaseModel):
    progress: List[Progress]
    result_status: bool
    solutions: List[QuboSolution]
    timing: SolverTiming


class QuboResponse(pydantic.BaseModel):
    qubo_solution: QuboSolutionList
    status: Literal["Done", "Deleted"]


class JobID(pydantic.BaseModel):
    job_id: str



import random
from .models import (
    Progress,
    QuboRequest,
    QuboResponse,
    QuboSolutionList,
    QuboSolution,
    SolverTiming,
)


def get_variable_number_keys(qubo_request: QuboRequest) -> list[str]:
    variable_number_set = set()

    if (
        "numbers" in qubo_request.fujitsuDA3.one_way_one_hot_groups
        and "numbers" in qubo_request.fujitsuDA3.two_way_one_hot_groups
    ):
        raise ValueError(
            "one_way_one_hot_groups and two_way_one_hot_groups cannot be specified at the same time."
        )

    if "numbers" in qubo_request.fujitsuDA3.one_way_one_hot_groups:
        num_one_way_one_hot_variables = sum(
            qubo_request.fujitsuDA3.one_way_one_hot_groups["numbers"]
        )
        variable_number_set.union(range(num_one_way_one_hot_variables))

    if "numbers" in qubo_request.fujitsuDA3.two_way_one_hot_groups:
        num_two_way_one_hot_variables = sum(
            qubo_request.fujitsuDA3.two_way_one_hot_groups["numbers"]
        )
        variable_number_set.union(range(num_two_way_one_hot_variables))

    if qubo_request.binary_polynomial is not None:
        for term in qubo_request.binary_polynomial.terms:
            for id in term.p:
                variable_number_set.add(id)

    if qubo_request.penalty_binary_polynomial is not None:
        for term in qubo_request.penalty_binary_polynomial.terms:
            for id in term.p:
                variable_number_set.add(id)

    if qubo_request.inequalities is not None:
        for inequality in qubo_request.inequalities:
            for term in inequality.terms:
                for id in term.p:
                    variable_number_set.add(id)

    return sorted([str(variable_number) for variable_number in variable_number_set])


def random_qubo_solution(variable_number_keys: list[str]):
    return QuboSolution(
        energy=-48.0,
        penalty_energy=0.0,
        frequency=1,
        configuration={
            variable_number_key: random.choice([True, False])
            for variable_number_key in variable_number_keys
        },
    )


def random_qubo_response(qubo_request: QuboRequest) -> QuboResponse:
    num_configurations = (
        qubo_request.fujitsuDA3.num_output_solution * qubo_request.fujitsuDA3.num_group
    )

    variable_number_keys = get_variable_number_keys(qubo_request)

    dummy_progress = [
        Progress(energy=0.0, penalty_energy=7.0, time=0.161),
        Progress(energy=-42.0, penalty_energy=0.0, time=0.355),
        Progress(energy=-45.0, penalty_energy=0.0, time=0.697),
        Progress(energy=-46.0, penalty_energy=0.0, time=1.109),
        Progress(energy=-47.0, penalty_energy=0.0, time=1.331),
        Progress(energy=-48.0, penalty_energy=0.0, time=1.629),
    ]
    dummy_result_status = True
    dummy_timing = SolverTiming(solve_time="11027", total_elapsed_time="11062")
    dummy_status = "Done"
    qubo_solutions = [
        random_qubo_solution(variable_number_keys) for _ in range(num_configurations)
    ]

    return QuboResponse(
        qubo_solution=QuboSolutionList(
            progress=dummy_progress,
            result_status=dummy_result_status,
            solutions=qubo_solutions,
            timing=dummy_timing,
        ),
        status=dummy_status,
    )



import unittest
from unittest.mock import patch, MagicMock, call

import json
import requests

from ommx_da4_adapter.client import DA4Client
from ommx_da4_adapter.exception import OMMXDA4AdapterError
from ommx_da4_adapter.models import (
    BinaryPolynomial,
    BinaryPolynomialTerm,
    FujitsuDA3Solver,
    JobID,
    QuboRequest,
)


class TestDA4Client(unittest.TestCase):
    def setUp(self):
        self.token = "test-token"
        self.url = "https://api.test.com/da"
        self.version = "v4"

        # Mock all HTTP methods of the requests module for the entire test
        patcher = patch("requests.get")
        self.mock_get = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = patch("requests.post")
        self.mock_post = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = patch("requests.delete")
        self.mock_delete = patcher.start()
        self.addCleanup(patcher.stop)

        # Set up default mock response
        mock_response = MagicMock()
        mock_response.json.return_value = {}
        self.mock_get.return_value = mock_response
        self.mock_post.return_value = mock_response
        self.mock_delete.return_value = mock_response

        self.client = DA4Client(token=self.token, url=self.url, version=self.version)
        self.expected_headers = {
            "accept": "application/json",
            "Content-Type": "application/json",
            "X-Api-Key": self.token,
        }

    def test_init_valid_version(self):
        client = DA4Client(token=self.token, url=self.url, version="v4")
        self.assertEqual(client._version, "v4")

        client = DA4Client(token=self.token, url=self.url, version="v3c")
        self.assertEqual(client._version, "v3c")

    @patch("requests.get")
    def test_get_success(self, mock_get):
        mock_response = MagicMock()
        mock_response.json.return_value = {"test": "data"}
        mock_get.return_value = mock_response

        result = self.client.get("/test/endpoint")

        mock_get.assert_called_once_with(
            self.url + "/test/endpoint", headers=self.expected_headers
        )
        self.assertEqual(result, {"test": "data"})

    @patch("requests.get")
    def test_get_error(self, mock_get):
        mock_response = MagicMock()
        mock_response.raise_for_status.side_effect = (
            requests.exceptions.RequestException
        )
        mock_response.text = "Error message"
        mock_get.return_value = mock_response

        with self.assertRaises(OMMXDA4AdapterError):
            self.client.get("/test/endpoint")

    @patch("requests.post")
    def test_post_success(self, mock_post):
        mock_response = MagicMock()
        mock_response.json.return_value = {"test": "data"}
        mock_post.return_value = mock_response

        test_data = json.dumps({"key": "value"})
        test_headers = {"header": "value"}

        result = self.client.post("/test/endpoint", test_data, test_headers)

        mock_post.assert_called_once_with(
            self.url + "/test/endpoint",
            headers=test_headers,
            data=test_data,
        )
        self.assertEqual(result, {"test": "data"})

    @patch("requests.post")
    def test_post_error(self, mock_post):
        mock_response = MagicMock()
        mock_response.raise_for_status.side_effect = (
            requests.exceptions.RequestException
        )
        mock_response.text = "Error message"
        mock_post.return_value = mock_response

        test_data = json.dumps({"key": "value"})
        test_headers = {"header": "value"}

        with self.assertRaises(OMMXDA4AdapterError):
            self.client.post("/test/endpoint", test_data, test_headers)

    @patch("requests.delete")
    def test_delete_success(self, mock_delete):
        mock_response = MagicMock()
        mock_response.json.return_value = {"test": "data"}
        mock_delete.return_value = mock_response

        result = self.client.delete("/test/endpoint")

        mock_delete.assert_called_once_with(
            self.url + "/test/endpoint", headers=self.expected_headers
        )
        self.assertEqual(result, {"test": "data"})

    @patch("requests.delete")
    def test_delete_error(self, mock_delete):
        mock_response = MagicMock()
        mock_response.raise_for_status.side_effect = (
            requests.exceptions.RequestException
        )
        mock_response.text = "Error message"
        mock_delete.return_value = mock_response

        with self.assertRaises(OMMXDA4AdapterError):
            self.client.delete("/test/endpoint")

    @patch.object(DA4Client, "post")
    def test_post_qubo_solve_without_blob(self, mock_post):
        mock_post.return_value = {"job_id": "test-job-id"}

        qubo_request = QuboRequest(
            fujitsuDA3=FujitsuDA3Solver(
                time_limit_sec=1,
                penalty_coef=10000,
                guidance_config={"1": False, "2": False, "4": False},
            ),
            binary_polynomial=BinaryPolynomial(
                terms=[
                    BinaryPolynomialTerm(c=2, p=[1, 2]),
                    BinaryPolynomialTerm(c=-4, p=[2, 4]),
                    BinaryPolynomialTerm(c=-3, p=[]),
                ]
            ),
        )

        result = self.client.post_qubo_solve(qubo_request)

        mock_post.assert_called_once_with(
            "/v4/async/qubo/solve",
            qubo_request.model_dump_json(exclude_none=True, by_alias=True),
            self.expected_headers,
        )
        self.assertEqual(result, "test-job-id")

    @patch.object(DA4Client, "post")
    def test_post_qubo_solve_with_blob(self, mock_post):
        mock_post.return_value = {"job_id": "test-job-id"}

        qubo_request = QuboRequest(
            fujitsuDA3=FujitsuDA3Solver(
                time_limit_sec=1,
                penalty_coef=10000,
                guidance_config={"1": False, "2": False, "4": False},
            ),
            binary_polynomial=BinaryPolynomial(
                terms=[
                    BinaryPolynomialTerm(c=2, p=[1, 2]),
                    BinaryPolynomialTerm(c=-4, p=[2, 4]),
                    BinaryPolynomialTerm(c=-3, p=[]),
                ]
            ),
        )
        blob_sas_token = "test-blob-token"
        blob_account_name = "test-account"

        expected_headers = {
            "accept": "application/json",
            "Content-Type": "application/json",
            "X-Api-Key": self.token,
            "X-Blob-Sas-Token": blob_sas_token,
            "X-Storage-Account-Name": blob_account_name,
        }

        result = self.client.post_qubo_solve(
            qubo_request,
            blob_sas_token=blob_sas_token,
            blob_account_name=blob_account_name,
        )

        mock_post.assert_called_once_with(
            "/v4/async/qubo/solve",
            qubo_request.model_dump_json(exclude_none=True, by_alias=True),
            expected_headers,
        )
        self.assertEqual(result, "test-job-id")

    @patch.object(DA4Client, "get")
    def test_get_jobs(self, mock_get):
        mock_get.return_value = {
            "job_status_list": [
                {
                    "job_id": "test-job-id-1",
                    "job_status": "Done",
                    "start_time": "2023-01-01T00:00:00Z",
                },
                {
                    "job_id": "test-job-id-2",
                    "job_status": "Running",
                    "start_time": "2023-01-01T00:01:00Z",
                },
            ]
        }

        result = self.client.get_jobs()

        mock_get.assert_called_once_with("/v4/async/jobs")
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0]["job_id"], "test-job-id-1")
        self.assertEqual(result[1]["job_status"], "Running")

    @patch.object(DA4Client, "get")
    def test_get_job_result(self, mock_get):
        mock_get.return_value = {
            "qubo_solution": "test_qubo_solution",
            "status": "Done",
        }
        job_id = "test-job-id"

        result = self.client.get_job_result(job_id)

        mock_get.assert_called_once_with("/v4/async/jobs/result/test-job-id")
        self.assertEqual(result["status"], "Done")

    @patch.object(DA4Client, "delete")
    def test_delete_job_result_success(self, mock_delete):
        mock_delete.return_value = {
            "qubo_solution": "test_qubo_solution",
            "status": "Deleted",
        }
        job_id = "test-job-id"

        result = self.client.delete_job_result(job_id)

        mock_delete.assert_called_once_with("/v4/async/jobs/result/test-job-id")
        self.assertTrue(result)

    @patch.object(DA4Client, "delete")
    def test_delete_job_result_failure(self, mock_delete):
        mock_delete.side_effect = OMMXDA4AdapterError("Failed to delete")
        job_id = "test-job-id"

        result = self.client.delete_job_result(job_id)

        mock_delete.assert_called_once_with("/v4/async/jobs/result/test-job-id")
        self.assertFalse(result)

    @patch.object(DA4Client, "post")
    def test_post_job_cancel_success(self, mock_post):
        mock_post.return_value = {"status": "Canceled"}
        job_id = "test-job-id"

        result = self.client.post_job_cancel(job_id)

        mock_post.assert_called_once_with(
            "/v4/async/jobs/cancel",
            JobID(job_id=job_id).model_dump_json(),
            self.expected_headers,
        )
        self.assertTrue(result)

    @patch.object(DA4Client, "post")
    def test_post_job_cancel_failure(self, mock_post):
        mock_post.side_effect = OMMXDA4AdapterError("Failed to cancel")
        job_id = "test-job-id"

        result = self.client.post_job_cancel(job_id)

        mock_post.assert_called_once_with(
            "/v4/async/jobs/cancel",
            JobID(job_id=job_id).model_dump_json(),
            self.expected_headers,
        )
        self.assertFalse(result)

    @patch.object(DA4Client, "get_jobs")
    @patch.object(DA4Client, "delete_job_result")
    def test_delete_all_jobs(self, mock_delete_job, mock_get_jobs):
        mock_get_jobs.return_value = [
            {"job_id": "test-job-id-1"},
            {"job_id": "test-job-id-2"},
        ]

        self.client.delete_all_jobs()

        mock_get_jobs.assert_called_once()
        mock_delete_job.assert_has_calls([call("test-job-id-1"), call("test-job-id-2")])

    @patch.object(DA4Client, "get_jobs")
    @patch.object(DA4Client, "post_job_cancel")
    def test_cancel_all_jobs(self, mock_cancel_job, mock_get_jobs):
        mock_get_jobs.return_value = [
            {"job_id": "test-job-id-1"},
            {"job_id": "test-job-id-2"},
        ]

        self.client.cancel_all_jobs()

        mock_get_jobs.assert_called_once()
        mock_cancel_job.assert_has_calls([call("test-job-id-1"), call("test-job-id-2")])

    @patch("time.sleep")
    @patch.object(DA4Client, "delete_job_result")
    @patch.object(DA4Client, "get_job_result")
    def test_fetch_job_result_success(self, mock_get_result, mock_delete, mock_sleep):
        job_id = "test-job-id"
        expected_result = {
            "status": "Done",
            "qubo_solution": {
                "progress": [{"energy": 0.0, "penalty_energy": 0.0, "time": 0.0}],
                "result_status": True,
                "solutions": [
                    {
                        "energy": 0.0,
                        "penalty_energy": 0.0,
                        "frequency": 1,
                        "configuration": {"1": True, "2": False},
                    }
                ],
                "timing": {"solve_time": "0.1s", "total_elapsed_time": "0.2s"},
            },
        }
        mock_get_result.return_value = expected_result

        result = self.client.fetch_job_result(job_id)

        mock_get_result.assert_called_once_with(job_id)
        mock_delete.assert_called_once_with(job_id)
        self.assertEqual(result.model_dump(), expected_result)

    @patch.object(DA4Client, "get_job_result")
    @patch("time.sleep")
    def test_fetch_job_result_running_then_done(self, mock_sleep, mock_get_result):
        job_id = "test-job-id"
        running_result = {"status": "Running"}
        done_result = {
            "status": "Done",
            "qubo_solution": {
                "progress": [{"energy": 0.0, "penalty_energy": 0.0, "time": 0.0}],
                "result_status": True,
                "solutions": [
                    {
                        "energy": 0.0,
                        "penalty_energy": 0.0,
                        "frequency": 1,
                        "configuration": {"1": True, "2": False},
                    }
                ],
                "timing": {"solve_time": "0.1s", "total_elapsed_time": "0.2s"},
            },
        }
        mock_get_result.side_effect = [running_result, done_result]

        result = self.client.fetch_job_result(job_id)

        self.assertEqual(mock_get_result.call_count, 2)
        mock_sleep.assert_called_once_with(self.client._solver_access_interval)
        self.assertEqual(result.model_dump(), done_result)

    @patch.object(DA4Client, "get_job_result")
    def test_fetch_job_result_error(self, mock_get_result):
        job_id = "test-job-id"
        error_result = {"status": "Error", "error": "test-error"}
        mock_get_result.return_value = error_result

        with self.assertRaises(OMMXDA4AdapterError):
            self.client.fetch_job_result(job_id)

    @patch.object(DA4Client, "get_job_result")
    def test_fetch_job_result_unknown_status(self, mock_get_result):
        job_id = "test-job-id"
        unknown_result = {"status": "Unknown"}
        mock_get_result.return_value = unknown_result

        with self.assertRaises(OMMXDA4AdapterError):
            self.client.fetch_job_result(job_id)

    @patch.object(DA4Client, "post_qubo_solve")
    @patch.object(DA4Client, "fetch_job_result")
    def test_sample(self, mock_fetch, mock_post_solve):
        job_id = "test-job-id"
        qubo_request = QuboRequest(
            fujitsuDA3=FujitsuDA3Solver(
                time_limit_sec=1,
                penalty_coef=10000,
                guidance_config={"1": False, "2": False, "4": False},
            ),
            binary_polynomial=BinaryPolynomial(
                terms=[
                    BinaryPolynomialTerm(c=2, p=[1, 2]),
                    BinaryPolynomialTerm(c=-4, p=[2, 4]),
                    BinaryPolynomialTerm(c=-3, p=[]),
                ]
            ),
        )
        mock_result = {
            "qubo_solution": {
                "progress": [{"energy": 0.0, "penalty_energy": 0.0, "time": 0.0}],
                "result_status": True,
                "solutions": [
                    {
                        "energy": 0.0,
                        "penalty_energy": 0.0,
                        "frequency": 1,
                        "configuration": {"1": True, "2": False},
                    }
                ],
                "timing": {"solve_time": "0.1s", "total_elapsed_time": "0.2s"},
            },
            "status": "Done",
        }

        mock_post_solve.return_value = job_id
        mock_fetch.return_value = mock_result

        result = self.client.sample(qubo_request)

        mock_post_solve.assert_called_once_with(
            qubo_request=qubo_request, blob_sas_token=None, blob_account_name=None
        )
        mock_fetch.assert_called_once_with(job_id)
        self.assertEqual(result, mock_result)

    @patch.object(DA4Client, "post_qubo_solve")
    @patch.object(DA4Client, "fetch_job_result")
    def test_sample_with_blob(self, mock_fetch, mock_post_solve):
        job_id = "test-job-id"
        qubo_request = QuboRequest(
            fujitsuDA3=FujitsuDA3Solver(
                time_limit_sec=1,
                penalty_coef=10000,
                guidance_config={"1": False, "2": False, "4": False},
            ),
            binary_polynomial=BinaryPolynomial(
                terms=[
                    BinaryPolynomialTerm(c=2, p=[1, 2]),
                    BinaryPolynomialTerm(c=-4, p=[2, 4]),
                    BinaryPolynomialTerm(c=-3, p=[]),
                ]
            ),
        )
        blob_sas_token = "test-blob-token"
        blob_account_name = "test-account"
        mock_result = {
            "qubo_solution": {
                "progress": [{"energy": 0.0, "penalty_energy": 0.0, "time": 0.0}],
                "result_status": True,
                "solutions": [
                    {
                        "energy": 0.0,
                        "penalty_energy": 0.0,
                        "frequency": 1,
                        "configuration": {"1": True, "2": False},
                    }
                ],
                "timing": {"solve_time": "0.1s", "total_elapsed_time": "0.2s"},
            },
            "status": "Done",
        }

        mock_post_solve.return_value = job_id
        mock_fetch.return_value = mock_result

        result = self.client.sample(
            qubo_request,
            blob_sas_token=blob_sas_token,
            blob_account_name=blob_account_name,
        )

        mock_post_solve.assert_called_once_with(
            qubo_request=qubo_request,
            blob_sas_token=blob_sas_token,
            blob_account_name=blob_account_name,
        )
        mock_fetch.assert_called_once_with(job_id)
        self.assertEqual(result, mock_result)


if __name__ == "__main__":
    unittest.main()



import pytest

from ommx.v1 import Constraint, DecisionVariable, Instance
from ommx.v1.constraint_hints_pb2 import ConstraintHints
from ommx.v1.one_hot_pb2 import OneHot

from ommx_da4_adapter import OMMXDA4Adapter
from ommx_da4_adapter.exception import OMMXDA4AdapterError
from ommx_da4_adapter.models import BinaryPolynomialTerm, Inequalities, QuboResponse


# Function to sort based on the length of binary polynomials
# Used to align the order (because the order may differ depending on the environment)
def sort_terms(terms):
    return sorted(terms, key=lambda term: (len(term.p), term.c))


@pytest.fixture
def instance_for_validation():
    x_1 = DecisionVariable.binary(id=0, name="x_1")
    x_2 = DecisionVariable.binary(id=1, name="x_2")

    objective = x_1 + x_2
    constraints = [x_1 * x_2 == 0]

    instance = Instance.from_components(
        decision_variables=[x_1, x_2],
        objective=objective,
        constraints=constraints,
        sense=Instance.MINIMIZE,
    )

    return instance


def test_time_limit_sec_validate(instance_for_validation):
    with pytest.raises(ValueError, match="`time_limit_sec` must be between 1 and 3600"):
        OMMXDA4Adapter(instance_for_validation, time_limit_sec=0)
    with pytest.raises(ValueError, match="`time_limit_sec` must be between 1 and 3600"):
        OMMXDA4Adapter(instance_for_validation, time_limit_sec=3601)


def test_target_energy_validate(instance_for_validation):
    with pytest.raises(
        ValueError, match=r"target_energy.*must be between \- 2\^126 and 2\^126"
    ):
        OMMXDA4Adapter(instance_for_validation, target_energy=-pow(2, 127))

    with pytest.raises(
        ValueError, match=r"target_energy.*must be between \- 2\^126 and 2\^126"
    ):
        OMMXDA4Adapter(instance_for_validation, target_energy=pow(2, 127))


def test_num_run_validate(instance_for_validation):
    with pytest.raises(ValueError, match="`num_run` must be between 1 and 1024"):
        OMMXDA4Adapter(instance_for_validation, num_run=0)
    with pytest.raises(ValueError, match="`num_run` must be between 1 and 1024"):
        OMMXDA4Adapter(instance_for_validation, num_run=1025)


def test_num_group_validate(instance_for_validation):
    with pytest.raises(ValueError, match="`num_group` must be between 1 and 16"):
        OMMXDA4Adapter(instance_for_validation, num_group=0)
    with pytest.raises(ValueError, match="`num_group` must be between 1 and 16"):
        OMMXDA4Adapter(instance_for_validation, num_group=17)


def test_num_output_solution_validate(instance_for_validation):
    with pytest.raises(
        ValueError, match="`num_output_solution` must be between 1 and 1024"
    ):
        OMMXDA4Adapter(instance_for_validation, num_output_solution=0)
    with pytest.raises(
        ValueError, match="`num_output_solution` must be between 1 and 1024"
    ):
        OMMXDA4Adapter(instance_for_validation, num_output_solution=1025)


def test_gs_level_validate(instance_for_validation):
    with pytest.raises(ValueError, match="`gs_level` must be between 0 and 100"):
        OMMXDA4Adapter(instance_for_validation, gs_level=-1)
    with pytest.raises(ValueError, match="`gs_level` must be between 0 and 100"):
        OMMXDA4Adapter(instance_for_validation, gs_level=101)


def test_gs_cutoff_validate(instance_for_validation):
    with pytest.raises(ValueError, match="`gs_cutoff` must be between 0 and 1000000"):
        OMMXDA4Adapter(instance_for_validation, gs_cutoff=-1)
    with pytest.raises(ValueError, match="`gs_cutoff` must be between 0 and 1000000"):
        OMMXDA4Adapter(instance_for_validation, gs_cutoff=1000001)


def test_one_hot_level_validate(instance_for_validation):
    with pytest.raises(ValueError, match="`one_hot_level` must be between 0 and 100"):
        OMMXDA4Adapter(instance_for_validation, one_hot_level=-1)
    with pytest.raises(ValueError, match="`one_hot_level` must be between 0 and 100"):
        OMMXDA4Adapter(instance_for_validation, one_hot_level=101)


def test_one_hot_cutoff_validate(instance_for_validation):
    with pytest.raises(
        ValueError, match="`one_hot_cutoff` must be between 0 and 1000000"
    ):
        OMMXDA4Adapter(instance_for_validation, one_hot_cutoff=-1)
    with pytest.raises(
        ValueError, match="`one_hot_cutoff` must be between 0 and 1000000"
    ):
        OMMXDA4Adapter(instance_for_validation, one_hot_cutoff=1000001)


def test_penalty_auto_mode(instance_for_validation):
    with pytest.raises(
        ValueError, match="`penalty_auto_mode` must be between 0 and 10000"
    ):
        OMMXDA4Adapter(instance_for_validation, penalty_auto_mode=-1)
    with pytest.raises(
        ValueError, match="`penalty_auto_mode` must be between 0 and 10000"
    ):
        OMMXDA4Adapter(instance_for_validation, penalty_auto_mode=10001)


def test_penalty_coef(instance_for_validation):
    with pytest.raises(
        ValueError, match="`penalty_coef` must be between 1 and 9223372036854775807"
    ):
        OMMXDA4Adapter(instance_for_validation, penalty_coef=0)
    with pytest.raises(
        ValueError, match="`penalty_coef` must be between 1 and 9223372036854775807"
    ):
        OMMXDA4Adapter(instance_for_validation, penalty_coef=9223372036854775808)


def test_penalty_inc_rate(instance_for_validation):
    with pytest.raises(
        ValueError, match="`penalty_inc_rate` must be between 100 and 200"
    ):
        OMMXDA4Adapter(instance_for_validation, penalty_inc_rate=99)
    with pytest.raises(
        ValueError, match="`penalty_inc_rate` must be between 100 and 200"
    ):
        OMMXDA4Adapter(instance_for_validation, penalty_inc_rate=201)


def test_max_penalty_coef(instance_for_validation):
    with pytest.raises(
        ValueError, match="`max_penalty_coef` must be between 0 and 9223372036854775807"
    ):
        OMMXDA4Adapter(instance_for_validation, max_penalty_coef=-1)
    with pytest.raises(
        ValueError, match="`max_penalty_coef` must be between 0 and 9223372036854775807"
    ):
        OMMXDA4Adapter(instance_for_validation, max_penalty_coef=9223372036854775808)


@pytest.fixture
def instance_for_integer_variable():
    x_1 = DecisionVariable.integer(id=0, name="x_1")
    x_2 = DecisionVariable.integer(id=1, name="x_2")

    objective = x_1 + x_2
    constraints = [x_1 * x_2 == 0]

    instance = Instance.from_components(
        decision_variables=[x_1, x_2],
        objective=objective,
        constraints=constraints,
        sense=Instance.MAXIMIZE,
    )

    return instance


def test_check_integer_variable(instance_for_integer_variable):
    with pytest.raises(
        OMMXDA4AdapterError,
        match=r"The decision variable must be binary: id .*",
    ):
        OMMXDA4Adapter(instance_for_integer_variable)


@pytest.fixture
def instance_for_continuous_variable():
    x_1 = DecisionVariable.continuous(id=0, name="x_1")
    x_2 = DecisionVariable.continuous(id=1, name="x_2")

    objective = x_1 + x_2
    constraints = [x_1 * x_2 == 0]

    instance = Instance.from_components(
        decision_variables=[x_1, x_2],
        objective=objective,
        constraints=constraints,
        sense=Instance.MAXIMIZE,
    )

    return instance


def test_check_continuous_variable(instance_for_continuous_variable):
    with pytest.raises(
        OMMXDA4AdapterError,
        match=r"The decision variable must be binary: id .*",
    ):
        OMMXDA4Adapter(instance_for_continuous_variable)


@pytest.fixture
def instance():
    x_1 = DecisionVariable.binary(id=0, name="x_1")
    x_2 = DecisionVariable.binary(id=1, name="x_2")
    x_3 = DecisionVariable.binary(id=2, name="x_3")

    # objective: 2x₁x₂x₃ + 3x₁x₂ + 4x₁ + 5x₂ + 6x₃ + 6
    # constraint_equality_1: x₁x₂x₃ == 0
    # constraint_equality_2: 3x₁x₂x₃ + 2x₁ + 5 == 0
    # constraint_inequality_1: 4x₁x₂x₃ ≤ 0
    # constraint_inequality_2: x₁x₂x₃ + 2x₃ + 3 ≤ 0

    objective = 2 * x_1 * x_2 * x_3 + 3 * x_1 * x_2 + 4 * x_1 + 5 * x_2 + 6 * x_3 + 6
    constraint_equality_1 = x_1 * x_2 * x_3 == 0
    constraint_equality_2 = 3 * x_1 * x_2 * x_3 + 2 * x_1 + 5 == 0
    constraint_inequality_1 = 4 * x_1 * x_2 * x_3 <= 0
    constraint_inequality_2 = x_1 * x_2 * x_3 + 2 * x_3 + 3 <= 0

    assert isinstance(constraint_inequality_1, Constraint)
    assert isinstance(constraint_inequality_2, Constraint)
    constraint_inequality_1.set_id(2)
    constraint_inequality_2.set_id(3)

    instance = Instance.from_components(
        decision_variables=[x_1, x_2, x_3],
        objective=objective,
        constraints=[
            constraint_equality_1,
            constraint_equality_2,
            constraint_inequality_1,
            constraint_inequality_2,
        ],
        sense=Instance.MINIMIZE,
    )

    return instance


def test_adapter_default_value(instance):
    adapter = OMMXDA4Adapter(instance)
    qubo_request = adapter.sampler_input

    # fujitsuDA3
    assert qubo_request.fujitsuDA3.time_limit_sec == 10
    assert qubo_request.fujitsuDA3.target_energy is None
    assert qubo_request.fujitsuDA3.num_run == 16
    assert qubo_request.fujitsuDA3.num_group == 1
    assert qubo_request.fujitsuDA3.num_output_solution == 5
    assert qubo_request.fujitsuDA3.gs_level == 5
    assert qubo_request.fujitsuDA3.gs_cutoff == 8000
    assert qubo_request.fujitsuDA3.one_hot_level == 3
    assert qubo_request.fujitsuDA3.one_hot_cutoff == 100
    assert qubo_request.fujitsuDA3.penalty_auto_mode == 1
    assert qubo_request.fujitsuDA3.penalty_coef == 1
    assert qubo_request.fujitsuDA3.penalty_inc_rate == 150
    assert qubo_request.fujitsuDA3.max_penalty_coef == 0
    assert qubo_request.fujitsuDA3.guidance_config is None
    assert qubo_request.fujitsuDA3.fixed_config is None

    assert adapter._inequalities_lambda is None

    assert qubo_request.bucket_name is None
    assert qubo_request.binary_polynomial_object_name is None
    assert qubo_request.penalty_binary_polynomial_object_name is None
    assert qubo_request.inequalities_object_name is None


def test_fujitsuDA3(instance):
    adapter = OMMXDA4Adapter(
        instance,
        time_limit_sec=10,
        target_energy=1,
        num_run=16,
        num_group=1,
        num_output_solution=5,
        gs_level=5,
        gs_cutoff=8000,
        one_hot_level=3,
        one_hot_cutoff=100,
        penalty_auto_mode=1,
        penalty_coef=1,
        penalty_inc_rate=150,
        max_penalty_coef=0,
        guidance_config={"description": True},
        fixed_config={"description": True},
    )
    qubo_request = adapter.sampler_input

    assert qubo_request.fujitsuDA3.time_limit_sec == 10
    assert qubo_request.fujitsuDA3.target_energy == 1
    assert qubo_request.fujitsuDA3.num_run == 16
    assert qubo_request.fujitsuDA3.num_group == 1
    assert qubo_request.fujitsuDA3.num_output_solution == 5
    assert qubo_request.fujitsuDA3.gs_level == 5
    assert qubo_request.fujitsuDA3.gs_cutoff == 8000
    assert qubo_request.fujitsuDA3.one_hot_level == 3
    assert qubo_request.fujitsuDA3.one_hot_cutoff == 100
    assert qubo_request.fujitsuDA3.internal_penalty == 0
    assert qubo_request.fujitsuDA3.penalty_auto_mode == 1
    assert qubo_request.fujitsuDA3.penalty_coef == 1
    assert qubo_request.fujitsuDA3.penalty_inc_rate == 150
    assert qubo_request.fujitsuDA3.max_penalty_coef == 0


def test_binary_polynomial(instance):
    adapter = OMMXDA4Adapter(instance)
    qubo_request = adapter.sampler_input

    # assert with manually calculated values
    # 2x₁x₂x₃ + 3x₁x₂ + 4x₁ + 5x₂ + 6x₃ + 6
    assert qubo_request.binary_polynomial is not None and sort_terms(
        qubo_request.binary_polynomial.terms
    ) == sort_terms(
        [
            BinaryPolynomialTerm(c=6.0, p=[]),
            BinaryPolynomialTerm(c=4.0, p=[0]),
            BinaryPolynomialTerm(c=3.0, p=[0, 1]),
            BinaryPolynomialTerm(c=2.0, p=[0, 1, 2]),
            BinaryPolynomialTerm(c=5.0, p=[1]),
            BinaryPolynomialTerm(c=6.0, p=[2]),
        ]
    )


def test_penalty_binary_polynomial(instance):
    adapter = OMMXDA4Adapter(instance)
    qubo_request = adapter.sampler_input

    # assert with manually calculated values
    # 52y₁y₂y₃ + 24y₁ + 25
    assert qubo_request.penalty_binary_polynomial is not None and sort_terms(
        qubo_request.penalty_binary_polynomial.terms
    ) == sort_terms(
        [
            BinaryPolynomialTerm(c=52.0, p=[0, 1, 2]),
            BinaryPolynomialTerm(c=24.0, p=[0]),
            BinaryPolynomialTerm(c=25.0, p=[]),
        ]
    )


def test_inequalities_lambda_default(instance):
    adapter = OMMXDA4Adapter(instance, inequalities_lambda={})
    qubo_request = adapter.sampler_input

    # if inequalities_lambda is None, failure
    assert qubo_request.inequalities is not None, "inequalities is None"

    for inequality in qubo_request.inequalities:
        assert inequality.lambda_ == 1


def test_inequalities_lambda(instance):
    adapter = OMMXDA4Adapter(instance, inequalities_lambda={2: 6, 3: 2})
    qubo_request = adapter.sampler_input

    # if inequalities_lambda is None, failure
    assert qubo_request.inequalities is not None, "inequalities is None"
    for inequality in qubo_request.inequalities:
        print(inequality)

    for inequality in qubo_request.inequalities:
        print(inequality.lambda_)
        if inequality.lambda_ == 2:
            assert inequality.lambda_ == 2
        elif inequality.lambda_ == 6:
            assert inequality.lambda_ == 6
        else:
            assert False


def test_inequalities(instance):
    adapter = OMMXDA4Adapter(instance)
    qubo_request = adapter.sampler_input

    # assert with manually calculated values
    # 4x₁x₂x₃ ≤ 0
    # x₁x₂x₃ + 2x₃ + 3 ≤ 0
    assert qubo_request.inequalities == [
        Inequalities(terms=[BinaryPolynomialTerm(c=4.0, p=[0, 1, 2])], lambda_=1),  # type: ignore
        Inequalities(
            terms=[
                BinaryPolynomialTerm(c=3.0, p=[]),
                BinaryPolynomialTerm(c=1.0, p=[0, 1, 2]),
                BinaryPolynomialTerm(c=2.0, p=[2]),
            ],
            lambda_=1,  # type: ignore
        ),
    ]


@pytest.fixture
def instance_for_MAXIMIZE():
    x_1 = DecisionVariable.binary(id=0, name="x_1")
    x_2 = DecisionVariable.binary(id=1, name="x_2")

    objective = x_1 + x_2
    constraints = [x_1 * x_2 == 0]

    instance = Instance.from_components(
        decision_variables=[x_1, x_2],
        objective=objective,
        constraints=constraints,
        sense=Instance.MAXIMIZE,
    )

    return instance


def test_binary_polynomial_for_MAXIMIZE(instance_for_MAXIMIZE):
    adapter = OMMXDA4Adapter(instance_for_MAXIMIZE)
    qubo_request = adapter.sampler_input

    # assert with manually calculated values
    # -y₁ - y₂
    assert qubo_request.binary_polynomial is not None and sort_terms(
        qubo_request.binary_polynomial.terms
    ) == sort_terms(
        [
            BinaryPolynomialTerm(c=-1.0, p=[0]),
            BinaryPolynomialTerm(c=-1.0, p=[1]),
            BinaryPolynomialTerm(c=-0.0, p=[]),
        ]
    )


@pytest.fixture
def instance_for_no_penalty_binary_polynomial():
    x_1 = DecisionVariable.binary(id=0, name="x_1")
    x_2 = DecisionVariable.binary(id=1, name="x_2")

    objective = x_1 + x_2
    constraints = [x_1 * x_2 <= 0]

    instance = Instance.from_components(
        decision_variables=[x_1, x_2],
        objective=objective,
        constraints=constraints,
        sense=Instance.MINIMIZE,
    )

    return instance


def test_no_penalty_binary_polynomial(instance_for_no_penalty_binary_polynomial):
    adapter = OMMXDA4Adapter(instance_for_no_penalty_binary_polynomial)
    qubo_request = adapter.sampler_input

    assert qubo_request.penalty_binary_polynomial is None


@pytest.fixture
def instance_for_no_inequalities():
    x_1 = DecisionVariable.binary(id=0, name="x_1")
    x_2 = DecisionVariable.binary(id=1, name="x_2")

    objective = x_1 + x_2
    constraints = [x_1 * x_2 == 0]

    instance = Instance.from_components(
        decision_variables=[x_1, x_2],
        objective=objective,
        constraints=constraints,
        sense=Instance.MINIMIZE,
    )

    return instance


def test_no_inequalities(instance_for_no_inequalities):
    adapter = OMMXDA4Adapter(instance_for_no_inequalities)
    qubo_request = adapter.sampler_input

    assert qubo_request.inequalities is None


@pytest.fixture
def instance_with_a_one_hot_constraint():
    x = [DecisionVariable.binary(id=i, name="x", subscripts=[i]) for i in range(5)]
    objective = sum(x[i] for i in range(5))
    constraint = [x[1] + x[2] + x[3] == 1]
    ommx_instance = Instance.from_components(
        decision_variables=x,
        objective=objective,
        constraints=constraint,
        sense=Instance.MAXIMIZE,
    )

    hints = ConstraintHints(one_hot_constraints=[OneHot(decision_variables=[1, 2, 3])])
    ommx_instance.raw.constraint_hints.MergeFrom(hints)

    for constraint in ommx_instance.get_constraints():
        constraint.set_id(0)

    return ommx_instance


def test_variable_map_with_a_one_hot_constraint(instance_with_a_one_hot_constraint):
    adapter = OMMXDA4Adapter(instance_with_a_one_hot_constraint)
    """
    for hint in hints.one_hot_constraints:
        print(hint.decision_variables)

    [1, 2, 3]
    """
    assert adapter._variable_map == {
        1: 0,
        2: 1,
        3: 2,
        0: 3,
        4: 4,
    }


def test_penalty_binary_polynomial_with_a_one_hot_constraint(
    instance_with_a_one_hot_constraint,
):
    adapter = OMMXDA4Adapter(instance_with_a_one_hot_constraint)
    qubo_request = adapter.sampler_input

    assert qubo_request.penalty_binary_polynomial is None


def test_one_way_one_hot_groups_with_a_one_hot_constraint(
    instance_with_a_one_hot_constraint,
):
    adapter = OMMXDA4Adapter(instance_with_a_one_hot_constraint)
    qubo_request = adapter.sampler_input

    assert qubo_request.fujitsuDA3.one_way_one_hot_groups is not None, (
        "one_way_one_hot_groups is None"
    )

    assert qubo_request.fujitsuDA3.one_way_one_hot_groups["numbers"] == [3]


def test_internal_penalty_with_a_one_hot_constraint(
    instance_with_a_one_hot_constraint,
):
    adapter = OMMXDA4Adapter(instance_with_a_one_hot_constraint)
    qubo_request = adapter.sampler_input

    assert qubo_request.fujitsuDA3.internal_penalty == 1


@pytest.fixture
def instance_with_various_constraints():
    x = [DecisionVariable.binary(id=i, name="x", subscripts=[i]) for i in range(10)]

    objective = sum(x[i] for i in range(10))
    constraint_1 = sum(x[i] for i in range(3)) == 1
    constraint_2 = sum(x[i] for i in range(3, 8)) == 1
    constraint_3 = sum(x[i] for i in range(8, 10)) <= 1

    assert isinstance(constraint_1, Constraint)
    assert isinstance(constraint_2, Constraint)
    assert isinstance(constraint_3, Constraint)
    constraint_1.set_id(0)
    constraint_2.set_id(1)
    constraint_3.set_id(2)

    ommx_instance = Instance.from_components(
        decision_variables=x,
        objective=objective,
        constraints=[constraint_1, constraint_2, constraint_3],
        sense=Instance.MAXIMIZE,
    )
    hints = ConstraintHints(
        one_hot_constraints=[
            OneHot(constraint_id=0, decision_variables=[0, 1, 2]),
            OneHot(constraint_id=1, decision_variables=[3, 4, 5, 6, 7]),
        ]
    )
    ommx_instance.raw.constraint_hints.MergeFrom(hints)

    return ommx_instance


def test_variable_map_with_various_constraints(instance_with_various_constraints):
    adapter = OMMXDA4Adapter(instance_with_various_constraints)
    """
    for hint in hints.one_hot_constraints:
        print(hint.decision_variables)

    [0, 1, 2]
    [3, 4, 5, 6, 7]
    """
    assert adapter._variable_map == {
        3: 0,
        4: 1,
        5: 2,
        6: 3,
        7: 4,
        0: 5,
        1: 6,
        2: 7,
        8: 8,
        9: 9,
    }


def test_one_way_one_hot_groups_with_various_constraints(
    instance_with_various_constraints,
):
    adapter = OMMXDA4Adapter(instance_with_various_constraints)
    qubo_request = adapter.sampler_input

    assert qubo_request.fujitsuDA3.one_way_one_hot_groups is not None, (
        "one_way_one_hot_groups is None"
    )

    assert qubo_request.fujitsuDA3.one_way_one_hot_groups["numbers"] == [5, 3]


@pytest.fixture
def instance_with_no_one_hot_constraint():
    x = [DecisionVariable.binary(id=i, name="x", subscripts=[i]) for i in range(5)]
    objective = sum(x[i] for i in range(5))
    constraint = [x[1] + x[2] + x[3] == 2]
    ommx_instance = Instance.from_components(
        decision_variables=x,
        objective=objective,
        constraints=constraint,
        sense=Instance.MAXIMIZE,
    )

    return ommx_instance


def test_variable_map_with_no_one_hot_constraint(instance_with_no_one_hot_constraint):
    adapter = OMMXDA4Adapter(instance_with_no_one_hot_constraint)
    """
    for hint in hints.one_hot_constraints:
        print(hint.decision_variables)

    []
    """
    assert adapter._variable_map == {
        0: 0,
        1: 1,
        2: 2,
        3: 3,
        4: 4,
    }


def test_one_way_one_hot_groups_with_no_one_hot_constraint(
    instance_with_no_one_hot_constraint,
):
    adapter = OMMXDA4Adapter(instance_with_no_one_hot_constraint)
    qubo_request = adapter.sampler_input

    assert qubo_request.fujitsuDA3.one_way_one_hot_groups is None


@pytest.fixture
def instance_with_duplicates():
    x = [DecisionVariable.binary(id=i, name="x", subscripts=[i]) for i in range(5)]
    objective = sum(x[i] for i in range(5))
    constraint_1 = sum(x[i] for i in range(2)) == 1
    constraint_2 = sum(x[i] for i in range(1, 4)) == 1

    assert isinstance(constraint_1, Constraint)
    assert isinstance(constraint_2, Constraint)
    constraint_1.set_id(0)
    constraint_2.set_id(1)

    ommx_instance = Instance.from_components(
        decision_variables=x,
        objective=objective,
        constraints=[constraint_1, constraint_2],
        sense=Instance.MAXIMIZE,
    )

    hints = ConstraintHints(
        one_hot_constraints=[
            OneHot(constraint_id=0, decision_variables=[0, 1]),
            OneHot(constraint_id=1, decision_variables=[1, 2, 3]),
        ]
    )
    ommx_instance.raw.constraint_hints.MergeFrom(hints)

    return ommx_instance


def test_variable_map_with_duplicates(instance_with_duplicates):
    adapter = OMMXDA4Adapter(instance_with_duplicates)
    """
    for hint in hints.one_hot_constraints:
        print(hint.decision_variables)

    [0, 1]
    [1, 2, 3]
    """
    # one hot constraints is only [1, 2, 3]
    assert adapter._variable_map == {
        1: 0,
        2: 1,
        3: 2,
        0: 3,
        4: 4,
    }


def test_one_way_one_hot_groups_with_duplicates(instance_with_duplicates):
    adapter = OMMXDA4Adapter(instance_with_duplicates)
    qubo_request = adapter.sampler_input

    assert qubo_request.fujitsuDA3.one_way_one_hot_groups is not None, (
        "one_way_one_hot_groups is None"
    )
    assert qubo_request.fujitsuDA3.one_way_one_hot_groups["numbers"] == [3]


def test_penalty_binary_polynomial_with_duplicates(instance_with_duplicates):
    adapter = OMMXDA4Adapter(instance_with_duplicates)
    qubo_request = adapter.sampler_input

    # assert with manually calculated values
    # (x_0 + x_1 - 1)^2
    assert qubo_request.penalty_binary_polynomial is not None and sort_terms(
        qubo_request.penalty_binary_polynomial.terms
    ) == sort_terms(
        [
            BinaryPolynomialTerm(c=-1.0, p=[3]),
            BinaryPolynomialTerm(c=2.0, p=[3, 0]),
            BinaryPolynomialTerm(c=-1.0, p=[0]),
            BinaryPolynomialTerm(c=1.0, p=[]),
        ]
    )


@pytest.fixture
def instance_for_sampleset():
    # Knapsack Problem
    v = [10, 13, 18, 31, 7, 15]
    w = [11, 25, 20, 35, 10, 33]
    W = 47
    N = len(v)

    x = [
        DecisionVariable.binary(
            id=i,
            name="x",
            subscripts=[i],
        )
        for i in range(N)
    ]

    objective = sum(v[i] * x[i] for i in range(N))

    constraint = sum(w[i] * x[i] for i in range(N)) - W <= 0
    assert isinstance(constraint, Constraint)

    instance = Instance.from_components(
        decision_variables=x,
        objective=objective,
        constraints=[constraint],
        sense=Instance.MAXIMIZE,
    )

    return instance


def test_decode_to_sampleset(instance_for_sampleset):
    adapter = OMMXDA4Adapter(instance_for_sampleset)

    # qubo_response from mock
    qubo_response = {
        "qubo_solution": {
            "progress": [
                {"energy": 0.0, "penalty_energy": 7.0, "time": 0.161},
                {"energy": -42.0, "penalty_energy": 0.0, "time": 0.355},
                {"energy": -45.0, "penalty_energy": 0.0, "time": 0.697},
                {"energy": -46.0, "penalty_energy": 0.0, "time": 1.109},
                {"energy": -47.0, "penalty_energy": 0.0, "time": 1.331},
                {"energy": -48.0, "penalty_energy": 0.0, "time": 1.629},
            ],
            "result_status": True,
            "solutions": [
                {
                    "energy": -48.0,
                    "penalty_energy": 0.0,
                    "frequency": 1,
                    "configuration": {
                        "0": False,
                        "1": True,
                        "2": False,
                        "3": True,
                        "4": True,
                        "5": False,
                    },
                },
                {
                    "energy": -48.0,
                    "penalty_energy": 0.0,
                    "frequency": 1,
                    "configuration": {
                        "0": False,
                        "1": True,
                        "2": True,
                        "3": True,
                        "4": True,
                        "5": False,
                    },
                },
                {
                    "energy": -48.0,
                    "penalty_energy": 0.0,
                    "frequency": 1,
                    "configuration": {
                        "0": False,
                        "1": False,
                        "2": False,
                        "3": True,
                        "4": True,
                        "5": False,
                    },
                },
                {
                    "energy": -48.0,
                    "penalty_energy": 0.0,
                    "frequency": 1,
                    "configuration": {
                        "0": True,
                        "1": False,
                        "2": True,
                        "3": False,
                        "4": False,
                        "5": True,
                    },
                },
                {
                    "energy": -48.0,
                    "penalty_energy": 0.0,
                    "frequency": 1,
                    "configuration": {
                        "0": False,
                        "1": False,
                        "2": False,
                        "3": False,
                        "4": False,
                        "5": True,
                    },
                },
            ],
            "timing": {"solve_time": "11027", "total_elapsed_time": "11062"},
        },
        "status": "Done",
    }

    sampleset = adapter.decode_to_sampleset(QuboResponse(**qubo_response))

    assert sampleset is not None
    assert len(sampleset.decision_variables) == 6
    assert len(sampleset.sample_ids) == 5


def test_sampele_without_token(instance_for_sampleset):
    with pytest.raises(
        OMMXDA4AdapterError,
        match="token is required. Please set the token to use the DA4 API.",
    ):
        OMMXDA4Adapter.sample(instance_for_sampleset)
